/* -*- c++ -*-
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: random.h,v 1.5 1994/08/19 18:38:31 bengt Exp $

#ifndef _random_h
#define _random_h

#ifdef __GNUG__
#pragma interface
#endif

#if !defined(BM) && !defined(PM)
#define INVPHI
#endif

#define MAX_SAVES 32
#define rand_deg 63
#define rand_sep 1

#include <limits.h>
#include <math.h>

#if 0
#if defined(__sparc) && defined(__GNUC__) && !defined(sqrt)
inline double static
sqrt(double x)
{
  register double result;
  asm("fsqrtd %1, %0" : "=f" (result) : "f" (x));
  return result;
}
#endif
#endif

class rng {
  long int *state;
  long int *fptr, *rptr;
  long int *end_ptr;
  long int randtbl[rand_deg];

  long int *saved_states[MAX_SAVES];
  unsigned int number_of_saved_states;
  
  long int int_threshold;
  int overhead;
  
#if 1
  double invmax;
  double invmax_2;
  double invmax_2_pi;
#else
  const double invmax	   = 0.46566128752457969241e-9; // 1/(2^31 - 1)
  const double invmax_2    = 0.93132257504915938482e-9; // 2/(2^31 - 1)
  const double invmax_2_pi = 0.29258361598967679948e-8; // 2 Pi /(2^31 - 1)
#endif
#ifndef INVPHI
  int one_left;
  double y;
#endif

  double sqr(double x) { return x*x; }
  unsigned long int sqr(unsigned long int x) { return x*x; }
  
  long int random()	// Returns an integer i; 0 <= i <= 2^31 - 1
    {
      *fptr += *rptr;
      long int i = (*fptr >> 1) & LONG_MAX;
      ++fptr;
      if (fptr >= end_ptr) {
	fptr = state;
	++rptr;
      } else {
	++rptr;
	if (rptr >= end_ptr)
	  rptr = state;
      }
      return i;
    }

  inline double uniform_0_2pi()    { return invmax_2_pi * random(); }
  inline double uniform_minus1_1() { return invmax_2 * random() - 1; }
      
#ifdef INVPHI

// Reference: Bratley, Fox, Schrage: A Guide to Simulation, p. 149f & 327

  double inv_phi_aux(double y)
    {
#ifndef __GNUC__
#define static
#endif
      static const double a0 = 0.322232431088;
      static const double a1 = 1.0;
      static const double a2 = 0.342242088547;
      static const double a3 = 0.0204231210245;
      static const double a4 = 0.453642210148e-4;

      static const double b0 = 0.0993484626060;
      static const double b1 = 0.588581570495;
      static const double b2 = 0.531103462366;
      static const double b3 = 0.103537752850;
      static const double b4 = 0.0038560700634;

      register double t = sqrt(-2*log(y));

      return t - (a0 + a1*t + a2*t*t + a3*t*t*t + a4*t*t*t*t)/
	(b0 + b1*t + b2*t*t + b3*t*t*t + b4*t*t*t*t);
    }

  inline double inv_phi(double y)
    {
#ifdef CHECK_VALID_INPUT
      if (y <= 0 || y >= 1) {
	errno = EINVAL;
	return NaN;
      }
#endif
      return y < 0.5 ? -inv_phi_aux(y) : inv_phi_aux(1-y);
    }
#endif

public:
  void srandom(long int x);
  void setstate(long int *arg_state);
  long int *getstate();
  int save_state(int index);  
  int save_state();
  int restore_state(int index);      
  rng();
  rng(unsigned long int n);
  ~rng();
  double uniform() 	{ return invmax*random(); }
  double normal()
    {
#ifdef INVPHI
      return inv_phi(uniform());
#else
      double u, v, x;
      double root;
#ifdef PM
      double w;
#endif


      if (one_left) {
	one_left = 0;
	return y;
      } else {
#ifdef PM
	do {
	  u = uniform_minus1_1();
	  v = uniform_minus1_1();
	} while ((w = sqr(u) + sqr(v)) > 1);
	root = sqrt(-2*log(w)/w);
	x = u * root;
	y = v * root;
#else
	u = uniform();
	v = uniform_0_2pi();
	root = sqrt(-2*log(u));
	x = root * cos(v);
	y = root * sin(v);
#endif
	one_left = 1;
	return x;
      }
#endif
    }

  void set_overhead(int n);
  double normal(int n)
    {
      register double x = 0;
      if (n) {
	for (int i = 1; i <= n; i++)
	  x += normal();
	return x/sqrt(n);
      } else
	return normal();
    }

  double normal_overhead()	{ return normal(overhead); }
  inline double normal(double m, double sigma) { return sigma*normal() + m; }  
  inline double normal(int n, double sigma)	{ return sigma*normal(n); }  
  void set_threshold(double p);
  inline int two_point()	{ return random() < int_threshold; }
  inline double two_point(double a, double b) 
    { return random() < int_threshold ? a : b; }
  inline double exponential(double lambda) { return -log(uniform())/lambda; }
};

#ifdef static
#undef static
#endif

#endif				// ! _random_h
